"""
Created on 9 Nov 2023
@author: daCunhaÉF and KraakmanYJ

Last edited by:
    daCunhaÉF on 04 Apr 2024
"""

import networkx as nx
import matplotlib.pyplot as plt
from networkx import single_source_all_shortest_paths
import numpy as np
import pandas as pd 
import re
import warnings
from scipy.signal import argrelextrema

### MAIN FUNCTIONS ###

def test():
    print('This file was imported')

def CRN2NET(file_path, directed=True, info=True):
    """
    Main function.

    It automatically creates a directed or undirected weighted species-species network 
    from chemical reaction equations set in .txt file.

    Node position doesn't have meaning.

    Edges encode the relationship in which one species is converted into 
    another through individuals chemical convertion as part of chemical reactions.

    Chemical reactions involving more than two species cannot be fully 
    represented by a single edge. While we currently assume that a set of edges 
    represents a reaction, we acknowledge the need for higher-order interactions, 
    such as a hyperedge, to better model chemical reactions. Consequently, the 
    implementation of higher-order network model is required.

    Parameters:
        file_path (str):                The txt file path with the chemical
                                        reaction equations, e.g. 'enzy_CRN_re_eqs.txt'.
                                       
        info (bool):                    Print additional information. If info=True 
                                        addtional information will be printed.

    Returns:
        species (list(str)):            Species and complexes.

        edges (list(tuple(str, str))):  Edges (source, target).

        eqs (list(list(str))):          Chemical reaction equations list.

    WARNING
        .txt file needs to:             Have one reversible equation for line 
                                        and use exacly the characters 
                                        ⇌ (U+21CC) and → (U+2192).
        Examples:
                                        A + B ⇌ [A.B] → 2 A
                                        A + C ⇌ [A.C] → A + D

    """
    ### CHEMICAL REACTION SYSTEM ###
    #open and reading the file
    file = open(file_path, "r", encoding='utf-8')
    file_content = file.read()
    file.close()

    ### SPECIES ASSIGNMENT ###
    #getting each equation
    CRN_eqs = file_content.splitlines()
    
    #Spliting reversible reactions into elementary reactions
    rever_eqs = [re.split(r'( ⇌ | → )', eq) for eq in CRN_eqs]
    eqs = rever2element_eqs(rever_eqs)
    eqs = clean_coefficients(eqs) #ignoring stoichiometric coefficients
    
    if directed: G = nx.DiGraph()
    else: G = nx.Graph()

    edges = []
    for eq in eqs:
        for species_in in eq[0].split(' + '): # IN/to nodes
            for species_out in eq[1].split(' + '): # OUT/from nodes
                if (species_in, species_out) in edges: # DIRECTED edges
                    G[species_in][species_out]["w"] += 1 # WEIGHTED edges
                else: 
                    ### COMPLEX NETWORK ###
                    G.add_edge(species_in, species_out, w=1) 
                    edges.append((species_in, species_out))
    if info==True:
        return G, list(G.edges(data=True))#, nx.adjacency_matrix(G, weight='w'), np.unique(eqs)
    else: return G

def build_crn_analysis_table(path_CRN_eqs, path_timeseries, start_end_time, fr_ap_cond, save=False, path_to_save='', print_=True):
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    #Time selection
    if start_end_time==[]: start_t, end_t = 0, None
    elif len(start_end_time)==1: start_t, end_t = start_end_time[0], None
    elif len(start_end_time)==2: start_t, end_t = start_end_time[0], start_end_time[1]

    skiprows = range(start_t) #lines to be ignored
    nrows = end_t - start_t + 1 #last line to be used
    
    #For code generalization, col_names should be an input, but for simplicity sake, here is just our case
    col_names = ['Time', 'Tr', 'Int-I', 'I', 'Tg', '<k_t(Tr + Int-I + I)>', 
                 'CC(Tg)','CC(Tr.Tg)',  'BC(Int-I)', 'BC(Tr.Pro-I)', 'BC(Ap.Int-I)']

    table = pd.DataFrame([], columns = col_names)
        
    G = CRN2NET(path_CRN_eqs, directed=True, info=False)#create the network
    for fr_ap in fr_ap_cond:
        fr, ap = fr_ap

        path_concentr = path_timeseries + '/CRN_('+str(fr)+'_fr;'+str(ap)+'_Ap)_flow.csv'
        path_bin = path_timeseries + '/CRN_('+str(fr)+' fr; '+str(ap)+' Ap)_flow_bin.csv'
        
        data_concentr = pd.read_csv(path_concentr, skiprows=range(1,start_t), nrows=nrows,
                                    delimiter=',', header=0, encoding='utf-8')
        data_bin = pd.read_csv(path_bin, skiprows=range(1, start_t), nrows=nrows, 
                                delimiter=',', header=0, encoding='utf-8')

        #Col 1: Time (h)
        table.loc[:, ('Time')] = data_bin.loc[:, 'Time']
        
        #Col 2, 3, 4, 5: concentration of Tr, Int-I, I, Tg
        table.loc[:, ('Tr', 'Int-I', 'I', 'Tg')] = data_concentr.loc[:, ('Tr', 'Int-I', 'I', 'Tg')] 
        
        #Col 6: <k_t(Tr + Int-I + I)>
        degrees = plot_measure_evolution(G, data_bin, measure='degree')
        numerator = degrees['Tr'] + degrees['Int-I'] + degrees['I']
        denominator = sum(dict(G.degree(['Tr', 'Int-I', 'I'], weight='w')).values())
        
        avr_degree_t = numerator/denominator
        table.loc[:, ('<k_t(Tr + Int-I + I)>')] = avr_degree_t

        #Col 7, 8: CC(Tg), CC(Tr.Tg)
        CCs = plot_measure_evolution(G, data_bin, measure='clustering')
        table.loc[:, 'CC(Tg)'] = CCs.loc[:, 'Tg']
        table.loc[:, 'CC(Tr.Tg)'] = CCs.loc[:, ('Tr.Tg')]

        #Col 9, 10, 11: BC(Int-I), BC(Tr.Pro-I), BC(Ap.Int-I)
        BCs = plot_measure_evolution(G, data_bin, measure='betweenness')
        table.loc[:, 'BC(Int-I)'] = BCs.loc[:, 'Int-I']
        table.loc[:, 'BC(Tr.Pro-I)'] = BCs.loc[:, 'Tr.Pro-I']
        table.loc[:, 'BC(Ap.Int-I)'] = BCs.loc[:, 'Ap.Int-I']
        
        if save: #default=False
            #Saving the table data for analysis
            path_save = path_to_save + 'analy_table_CRN_('+str(fr)+' fr; '+str(ap)+' Ap)_flow.csv'
            table.to_csv(path_save, index=False)
            print('Done:', path_save)
        if print_: print(table)


def G_snap(G, bin_data, t_snap, draw=True, edge_list=True, list_to_del=['W']):
    """
    Generates a new complex network, G_snap, inherited from the species-species
    network resulting from chemical reactions at time moment t_snap in the 
    species evolution time-series obtained from the system simulation.

    Parameters:
        G (nx.DiGraph):                 Original species-species network.

        bin_data (pd.DataFrame):       Species evolution matrix with shape:
                                        #samples x #species 

        t_snap (int):                   The simulation sample moment that will 
                                        be encoded into a new species network.
                                        E.g. with 1 sample every 0.01h, the
                                        t_snap=100 is at 1h real time.

        draw (bool, optional):          Default is True, generates a visual 
                                        representation of the network.

        edge_list (bool, optional):     Default is True, prints the network edge list.

        list_to_del (list, optional):   Element list to be excluded from the network.

    Returns:
        G_t(nx.DiGraph):                G at moment t_snap.

        list(tuple(str, str, dict)):    [(source, target, {weight:int})].
        
    """      

    G_t = G.copy()      #Deep copy to create a new object, G_t, with the  
                        #same information of G, but in different memory location. 
                        #Same network, different objects, changing one 
                        #doesn't change the other

    filter_dict = time_filter(G_t, bin_data, t_snap)    #binary time-serie 
                                                        #informing which edges  
                                                        #of the original crn_net 
                                                        #will be deleted

    edges_to_del = [(elem, nei) for elem in filter_dict if filter_dict[elem] == 0 \
                    for nei in G_t.neighbors(elem)] #filter application
    G_t.remove_edges_from(edges_to_del) #encodes G at moment t_snap

    if draw: draw_crn_net(G_t)
    if edge_list: print(pd.DataFrame(list(G_t.edges(data=True)),
                                    columns=['Source', 'Target', 'Attributes']))
        
    return G_t, list(G_t.edges(data=True))

def G_snap_with_dict(G, filter_dict, draw=True, edge_list=True):
  
    """
    G_snap test function with the binary data enconded in the filter_dict.
    It's not used in the process, but it can be useful to check specific G_snap 

    Ex: filter_dict = {"Tr":1,
                        "Tg":1,
                        "[Tr.Tg]":1,
                        "[Tr.a1a2I]":1,
                        "a1a2I":1,
                        "a2I":1,
                        "[Ap.a2I]":1,
                        "Ap":1,
                        "I":1,
                        "W":1}

    Parameters:
        G (nx.DiGraph):                 Original species complex network.

        filter_dict (dict):             dict to incorporate the binarized values 
                                        of G_snap function manually.

        draw (bool, optional):          If True, generates a visual 
                                        representation of the network.

        edge_list (bool, optional):     If True, prints the network edge list.


    Returns:
        G_snap(nx.DiGraph):             G (specie-species network) at t_snap.

        list(tuple(str, str, dict)):    [(source, target, {weight:int})].
    """ 

    G_snap = G.copy()   #Deep copy to create a new object, G_snap, with the  
                        #same information of G, but in different memory location. 
                        #Same network, different objects, changing one 
                        #doesn't change the other


    edges_to_del = [(elem, nei) for elem in filter_dict if filter_dict[elem] == 0 \
                    for nei in G_snap.neighbors(elem)] #filter application
    G_snap.remove_edges_from(edges_to_del) #encodes G at moment t_snap

    if draw: draw_crn_net(G_snap)
    if edge_list: print(pd.DataFrame(list(G_snap.edges(data=True)), 
                                    columns=['Source', 'Target', 'Attributes']))
        
    return G_snap, list(G_snap.edges(data=True))

def characterizing(G, measure):
    """
    Characterizes the nodes in a given network using a specified measure.
    
    Parameters:
        - G (nx.DiGraph):       The input network for node characterization.

        - measure (str):        The centrality measure to be applied.  
                                Currently:'degree', 'betweenness', 'clustering'.

    Returns:
        - char_dict (dict):     A dictionary containing node identifiers as keys 
                                and their associated centrality values based on
                                the chosen measure.

    Supported Centrality Measures:
        - 'degree':             Computes the node degree centrality, 
                                optionally weighted.
        - 'betweenness':        Computes the node betweenness centrality.
        
        - 'clustering':         Computes the node local clustering coefficient.
    """
    if measure == 'degree': return dict(nx.degree(G, weight='w'))
    elif measure == 'betweenness': return betweenness_centrality(G)
    elif measure == 'clustering': return clustering_local(G)

def draw_crn_net(G, weighted=True, save_path=False, title=False, pos=False, specific_sys=False, transparent=False):
    """
    It plots a directed and weighted species-species network of CRN.
    This is not a general function, scripted for our case. But it can
    be modified for other situations.
    """
    fig = plt.figure(figsize=(6.5, 6.5))
    
    #Selection of nodes to be used on our two network verions, under- and out-of-flow
    if specific_sys == 'enzy_oscil_flow':
        rename_map = {'Tg':'1','Ap':'2','Pro-I':'3','Tr':'4',
                      'Int-I':'5','I':'6', 'Tr.Tg':'7','Tr.Pro-I':'8',
                      'Ap.Int-I':'9', 'P':'10', 'S':'11', 'W':'12'}
        G = nx.relabel_nodes(G, rename_map)
        
    elif specific_sys == 'enzy_oscil_batch':
        rename_map = {'Tg':'1','Ap':'2','Pro-I':'3','Tr':'4',
                      'Int-I':'5','I':'6', 'Tr.Tg':'7','Tr.Pro-I':'8',
                      'Ap.Int-I':'9', 'P':'10'}
        G = nx.relabel_nodes(G, rename_map)
    edge_boldness = 2
    
    # Extract weights from the edges
    if weighted: weights = [edge[2]['w']*edge_boldness for edge in G.edges(data=True)]
    else: weights = 1*edge_boldness

    if pos is False: #If True, pos should be inputed. It is used to draw all G versions
                     #with nodes having the same position throughout G_snaps. However, it 
                     #seems to not work well, it can be seeing in the evolution of network
                     #structure (net_evol_single_oscil.gif in Extended Materials), where the
                     #node position change in a specific moment.
        pos = nx.circular_layout(G)
        
    # Draw the nodes
    nx.draw_networkx_nodes(G, pos, node_size=500, node_color='black')
    
    # Draw the edges with weights
    edge_boldness = 50
    nx.draw_networkx_edges(G, pos,
                          #connectionstyle="arc3, rad=0.1", \
                           arrowstyle='-|>',arrowsize=15,
                           width=weights*edge_boldness, edge_color='gray')
    
    # Add node labels
    nx.draw_networkx_labels(G, pos, font_color='white', font_weight='bold',
                           horizontalalignment='center', verticalalignment='center')
    
    plt.axis('off')  # Turn off the axis
    if title is not False: plt.title(title)
    plt.show()
    
    #Settings to save the network figures
    if save_path is not False:
        if transparent:
            fig.patch.set_alpha(0)
            fig.savefig(save_path +'.png', transparent=True)
        elif transparent is False:
            fig.savefig(save_path +'.png')

def plot_measure_evolution(G, evol_data, measure, species2show=[], marker='', plot=False):
    """
    Plot the measure evolution over time.

    Parameters:
        G (nx.DiGraph):                 Weighted, directed network G (no double edges)

        evol_data (dict):               Species evolution and respective time

        measure (str):                  The centrality measure to be plotted over time. 
                                        'degree', 'betweenness', 'clustering'.

        species2show (list, optional):  List of species to be shown on the plot. 
                                        If empty, it plots all network species.

        marker (str, optional):         Marker style for the plotted data. 
                                        If empty, a default markers will be used.
        
    Returns:
        measure_evol (pd.DataFrame):    DataFrame with measure values over time.
        
    PLOT:
        Measure time-series plot

    Example:
        species2show = ['Tr', '[Tr.Tg]', 'Tg', '[Tr.a1a2I]', 'a1a2I', 'a2I', 
                        'Ap', '[Ap.a2I]', 'I']
        measure = 'degree' # or 'betweenness' and 'clustering'
        measure_evol = plot_measure_evolution(G, evol_data, measure=measure, 
                                              species2show=species2show)
    """

    measure_evol = pd.DataFrame()
    time = evol_data['Time']

    for t in np.arange(len(time)):
        G_t, edge_list = G_snap(G, evol_data, t_snap=t, draw=False, edge_list=False)
        measure_values = pd.DataFrame(characterizing(G_t, measure), index=[t])
        measure_evol = pd.concat([measure_evol, measure_values], ignore_index=True)

    if plot:
        if species2show==[]: species2show=list(G.nodes())
        markers = ['s','D','o','<','>','v','^','x','*','H','8']
        if marker == '': marker=markers
        else: marker = marker*len(species2show)
            
        plt.figure(figsize=(10, 6))
        for m, specie in enumerate(species2show):
            plt.plot(time, measure_evol[specie], label=specie, marker=marker[m])
    
        plt.xlabel('Time (h)')
        plt.ylabel(measure)
        plt.title(measure + ' Evolution')
        plt.legend()
        plt.grid()
        plt.show()
    return measure_evol

def oscillation_period_detection(timeserie):
    """
    Responsible to detect and select the full period of single oscillations 
    of an oscillatory time series
    """
    valleys_id = argrelextrema(timeserie, np.less)[0]
      
    # Periods
    Ts_id = np.diff(valleys_id)#T0 = valley1 - valley0
    Ts_id_moda, Ts_id_moda_freq = np.unique(Ts_id, return_counts=True)

    for T_id_moda in Ts_id_moda: #more frequent even period
        if T_id_moda % 2 == 0:
            T_id_moda_len = T_id_moda
            break

    oscil_id_dict = {f'p{i}': [p, p+T_id_moda_len] for i, p in enumerate(valleys_id)}   
    return pd.DataFrame(oscil_id_dict), valleys_id

def all_oscillations_1period(data, timeserie_baseline, N_valleys=5):#, save_table=False):
    """
    Responsible for 'normalize' multiple oscillations into a single period.
    """
    #Selecting the index of each period of N_valleys oscillations
    oscil_id_dict, valleys_id = oscillation_period_detection(np.array(timeserie_baseline))
    valley_selected = [f'p{v}' for v in range(N_valleys)]

    #id for each oscillation
    oscil_limits = oscil_id_dict[valley_selected]
    period_size = oscil_limits['p0'][1]-oscil_limits['p0'][0]
    period = np.linspace(0, 1, period_size)

    #Crete the final table with first col as 'time'/1-period
    table_final = pd.DataFrame({'Period': period})
    data['Tr'] /= np.max(data['Tr']) #normalizing Tr values
    data['Int-I'] /= np.max(data['Int-I'])
    data['I'] /= np.max(data['I'])
    data['Tg'] /= np.max(data['Tg'])

    #Selection of data point respective for the period of single oscillation 
    #for multiples oscillations, and avarege and std of them.
    for col in data.columns[1:]: #avoiding time/first col
        aux = []
        for v in valley_selected:
            oscil_start, oscil_end = oscil_limits[v]
            aux.append(data[col][oscil_start:oscil_end])
            
        table_final[f'{col}_mean'] = np.mean(aux,axis=0)
        table_final[f'{col}_std'] = np.std(aux,axis=0)

    return table_final

def start_end_id(timeseries):
    """
    Detection of start and end points of oscillations
    """
    start_id = next((i for i, val in enumerate(timeseries) if val != 0), None)
    end_id = next((i for i, val in enumerate(reversed(timeseries[start_id:])) if val == 0), None)

    if end_id is not None:
        end_id += start_id 

    return start_id, end_id #[start_id, end_id[

#################### COMPLEMENTARY FUNCTIONS ########################


def rever2element_eqs(rever_eqs):
    """
    It splits the reversible equations into elementary equations.

    Example:                A + C ⇌ [A.C] → A + D turns into:
                     i.     A + C → [A.C]
                     ii.    [A.C] → A + C  
                     iii.   [A.C] → A + D
    """
    result = []    
    for item in rever_eqs:
        for i in range(len(item)):
            if item[i] == " ⇌ ":
                result.append([item[i-1],item[i+1]])
                result.append([item[i+1],item[i-1]])
            elif item[i] == " → ":
                result.append([item[i-1],item[i+1]])

    return result
    
def time_filter(G, bin_data, t_snap, list_to_del=['W', 'S', 'P']):
    """
    Responsible for get the binary evol_data at time t_snap and return a
    dictonary of species with binary value at t_snap
    """
    species = list(G.nodes())
    
    species = del_species(species, list_to_del)
    filter_dict = {specie:bin_data.at[t_snap, specie] for specie in species}
    return filter_dict

def del_species(species, species2del):
    """
    Responsible for deleting problematic species (species not in binary data):
        "W" is standard because it is a combination of species, not a sigle one
        to be detected;
    """

    for s_del in species2del:
        species.remove(s_del)
    return species

def clean_coefficients(input_list):
    """
    It deletes the stoichiometric coefficients.

        Example:    [A.B] → 2 A turns into [A.B] → A
    """
    
    cleaned_list = []
    for pair in input_list:
        cleaned_pair = []
        
        for term in pair:
            cleaned_term = re.sub(r'\d+\s+', '', term)
            cleaned_pair.append(cleaned_term)
        
        cleaned_list.append(cleaned_pair)
    return cleaned_list

def clustering_local_single(G, node):
    """
    Computes the local clustering coefficient for individual nodes in the network

    Parameters:
    G (nx.DiGraph): weighted, directed network G (no double edges) node for which 
                    the CC must be computed.

    node (str):     The node for which the local CC is to be computed.

    Returns:    
        CC (float): Local clustering coefficient for node x, considering its 
                    neighbors y and z:
        
                    CC(x) = (#triangles x-y-z, weight is considered 1 for each edge)/
                            (#wedges z-x-y, weight is considered 1 for each edge).
    
    Created on Mon Nov  6 09:39:05 2023
    @author: KraakmanYJ

    Edited by:
        daCunhaÉF on 09 Nov 2023
        daCunhaÉF on 10 Nov 2023
    """
    wedges=0
    triangles=0
    for pred in set(G.pred[node]) - {node}: #iterate over predecessors, exclude node itself
        for succ in set(G.succ[node]) - {node}: #iterate over successors, exclude node itself
            if pred != succ: #if pred is unequal to succ then a wedge is found
                wedges += 1
                if succ in set(G.pred[pred]): #wedge closes into a triangle
                    triangles += 1
    if wedges == 0: return 0
    else: return triangles/wedges

def clustering_local(G):
    """
    Computes the local clustering coefficient for each node in the graph

    Parameters:
        G (nx.DiGraph):         weighted, directed graph G (no double edges)
                                node for which the CC must be computed.

    Returns:                
        CC (dict):              Dictionary with local clustering coefficients 
                                for each node.

    Created on Mon Nov  6 09:39:05 2023
    @author: KraakmanYJ
    Edited by:
        daCunhaÉF on 09 Nov 2023
        daCunhaÉF on 10 Nov 2023
    """
    CC={}
    for node in G.nodes:
        CC[node] = clustering_local_single(G, node)
    return CC

def betweenness_centrality(G):
    """
    Computes the betweenness centrality values of all nodes in G, based on the 
    unweighted shortest paths. Edges with weight > 1 are considered multiple 
    edges, resulting in multiple shortest paths. Each such shortest path is 
    counted separately. 
    
    Maths:
        BC(i) = 1/(|P-p(i)|) \sum_{s,t \in P - p(i)} \sigma_{s,t}(v) / \sigma_{s,t},    (1)
        
        where
        P = {(v,w): v,w \in V and a path from v to w exists},
        p(i) = {(v,w) : (v,w) \in P and v=i or w=i},
        \sigma_{s,t}(v) = nr of shortest paths from s to t passing through v,
        \sigma_{s,t} = nr of shortest paths from s to t.
        
        Extra info:
            P-p(i) represents the node pairs (excluding i) between
                which a path exists. Dividing 1 by the cardinality of this set
                gives us the normalizing factor: if all shortest paths between 
                those pairs of nodes pass through i then BC(i) = 1. If none do, 
                then BC(i) = 0.
                NOTE: this normalizing factor can be different for every node.

    Parameters:
        G (nx.DiGraph):     weighted, directed graph G (no double edges)

    Returns: 
        BC (dict):          Dictionary with betweenness centrality

    Created on Mon Nov  6 11:14:41 2023
    @author: KraakmanYJ
    Edited by:
        daCunhaÉF on 09 Nov 2023
        daCunhaÉF on 10 Nov 2023
        KraakmanYJ on 12 Feb 2024
    """
    sum_per_node=dict.fromkeys(G, 0) #will produce the value of the sum in (1) for every node
    connected_pairs = 0 #nr of connected node pairs (=|P|), initial value 
    connected_pairs_including_node = dict.fromkeys(G, 0) #nr of node pairs, including specific node (=|p(i)|)
    for source in G.nodes: #pick node s
        shortest_paths = single_source_all_shortest_paths(G,source) #shortest paths starting from source
        for (target,paths_to_this_target) in shortest_paths: #all shortest paths to a specific target
            if target != source:  
                sigma = 0 #nr of shortest paths from s to t (=\sigma_{s,t}), initial value
                sigma_via_node = {} #nr of shortest paths from s to t via specific nodes (=\sigma_{s,t}(v))
                connected_pairs += 1 #source and target are a connected pair
                connected_pairs_including_node[source] += 1
                connected_pairs_including_node[target] += 1                
                for path in paths_to_this_target: #one path from source to target
                    #check the weights of the edges in this paths: if weight >1 then count the path multiple times.
                    number_of_paths = weights_product(G,path)
                    sigma += number_of_paths #number of paths from s to t increases
                    for between_node in path[1:-1]: #nodes in path that are not source or target node
                        #increase sigma_via_node (=sigma_{s,t}(v)) for these nodes
                        if between_node in sigma_via_node:
                            sigma_via_node[between_node] += number_of_paths
                        else:
                            sigma_via_node[between_node] = number_of_paths
                for node in sigma_via_node: #for all nodes that were in between paths from this source to target
                    sum_per_node[node] += (sigma_via_node[node])/sigma
    BC = {} #betweenness centrality value
    for node in G.nodes: #normalize the sum_per_node value
        BC[node] = (sum_per_node[node])/(connected_pairs - connected_pairs_including_node[node]) 
    return BC

def weights_product(G,path):
    """
    Multiplies the weights of the edges in the path.
    
    Parameters:
        G (nx.DiGraph):           Weighted, directed graph G (no double edges)

    Returns: 
        product (int):            Weight of an edge

    Created on Mon Nov  6 11:14:41 2023
    @author: KraakmanYJ
    Edited by:
        daCunhaÉF on 09 Nov 2023
    """
    product = 1
    for tail_index in range(len(path)-1):
        #now we want to find the weight of edge (tail,head),
        #where tail=path[tail_index] and head=path[tail_index+1]
        tail=path[tail_index]
        head=path[tail_index+1]
        weight = G.get_edge_data(tail,head)['w']
        product *= weight
    return product





