# -*- coding: utf-8 -*-
"""
Created on Wed Aug  9 11:41:35 2023

@author: KraakmanYJ

This code can be used to uniformly sample directed graphs that have the same 
degree sequence as some initial directed graph G. One needs to choose the 
graph space to sample from:
    - (dis)allow self-loops: if self-loops could exist in your application, then
            the samples should be allowed self-loops as well
    - stub- or vertex-labeled space: depending on your application, one of these
            spaces applies. See Fosdick et al. 'Configuration random graph 
            models with fixed degree sequences' 2018
                            
main function: randomize()
"""

import matplotlib.pyplot as plt
import networkx as nx
import random
import copy

"""
RANDOMIZATION

Randomization is performed using edge swapping. 
1) two edges are picked uniformly at random
2) the heads of the edges are swapped
3) if the swap is not valid (e.g., results in a self-loop while those are
        picked as disallowed) then the swap is undone. 
4) when sampling from the vertex-labeled space, acceptance probabilities are
        used for every swap. If not accepted, then the swap is undone. 
"""

def randomize(G, self_loops=True, labeled='vertex', mixingTime=None):
    """
    function that generates a uniformly random sample with the same degree 
        sequence as some input graph.
    input: 
        1) multidigraph G, 
        2) self_loops: True if self-loops are allowed, False else
        3) labeled: 'stub' or 'vertex', if sampling from stub-labeled or 
            vertex-labeled graph space, respectively.
        4) mixingTime: number of swaps done. Default is #edges*50.
    returns: randomized multidigraph, with the same degree sequence as G. 
            multi-edges are allowed, self-loops are allowed only if self_loops=True
    """
    if mixingTime==None:
        mixingTime = len(G.edges)*50
    randomG = copy.deepcopy(G) #graph that will be randomized
    for i in range(mixingTime):
        randomG = swap_edges(randomG,self_loops,labeled) #perform a swap
    return randomG
        
def swap_edges(G,self_loops=False, labeled='vertex'):
    """
    function that performs 1 swap in a multidigraph. If the swap results in a
        self-loops when these are not allowed, then the swap is undone. When
        sampling from a vertex-labeled space, an acceptance probability is used
        for the swap.
    input: 
        1) unweighted multidigraph
        2) self_loops: True if self-loops are allowed, False else
        3) labeled: 'stub' or 'vertex', if sampling from stub-labeled or 
            vertex-labeled graph space, respectively.
    returns:
        the multidigraph with two edges swapped (or the original multidigraph
            if the swap was rejected)
                             
    """
    edges=random.sample(list(G.edges),k=2) #sample two edges uniformly at random
    new_edges = [(edges[0][0],edges[1][1]),(edges[1][0],edges[0][1])] #swap the heads
    if not self_loops: #if self-loops are not allowed, check for them
        for i in range(len(new_edges)): #check for self-loops in new_edges
            tail,head = new_edges[i] #new edge that is checked
            if tail == head: #self-loop
                return G #don't perform the swap
    if labeled == 'vertex':
        #when sampling from a vertex labeled space, use acceptance probabilities 
        if random.uniform(0,1) > acceptanceProbability(G,edges): #acceptance probabiliy is NOT met
            return G #don't perform the swap
    #the swap is accepted, so G will be adjusted accordingly.
    G.remove_edges_from(map(lambda x: (x[0],x[1]), edges)) #remove edges that are swapped
    G.add_edges_from(map(lambda x: (x[0],x[1]), new_edges)) #add edges resulting from swap
    return G

def acceptanceProbability(G,sampled_edges):
    """
    computes the acceptance probability of a swap, when sampling from a 
        vertex-labeled graph space. For swapping e_1 and e_2, the acceptance
        probability is computed as follows:
        (multiplicity of e_1 in G.edges * multiplicity of e_2 in G.edges)^(-1)
    input:
        1) multidigraph G
        2) two edges that are selected to swap.
    returns:
        the acceptance probability of the swap. (float)
            
    """
    multiplicity = {} #will count the multiplicity of e_1 and e_2
    for sampled_edge in sampled_edges:
        multiplicity[sampled_edge]=0 #initialize multiplicities as 0
    for edge in G.edges: #check all edges
        for sampled_edge in sampled_edges: #for e_1 and e_2
           if edge[0] == sampled_edge[0] and edge[1] == sampled_edge[1]: #edge equals sampled edge
                multiplicity[sampled_edge] += 1
    acceptanceProbability = 1 #initializing the acceptance probability
    for sampled_edge in sampled_edges: #for e_1 and e_2
        acceptanceProbability /= multiplicity[sampled_edge]
    return acceptanceProbability



    
        
        
        
    

