# -*- coding: utf-8 -*-
"""
Created on Wed Aug  9 11:41:35 2023

@author: KraakmanYJ

This code can be used to obtain a null model for certain graph measures.
Main function: create_data
"""

import matplotlib.pyplot as plt
import networkx as nx
import pandas as pd
import seaborn as sns
import copy

from ConfigModel_EdgeSwapping_multidigraph import randomize
from CRN_network_tools import clustering_local, betweenness_centrality



def create_data(G,nrSamples, measures, self_loops=True, labeled='vertex'):
    """
    Function that creates a number of random network samples, applies a list of
        measures on each sample, and stores the obtained data as lists in a 
        dictionary.
    
    input: 
        1) weighted digraph G
        2) nr of samples to use for the null model (int)
        3) list of measures to use for the null model (list of strings). 
            NOTE: only the measures 'betweenness_centrality' and 'clustering'
                    are implemented in this code. Other measures are not
                    recognized.
        4) self_loops: True if self-loops are allowed (Boolean)
        5) labeled: vertex of stub (string)
    returns:  
        dictionary. Every measure in the list 'measures' is a key. The value
            corresponding to each such key is a dictionary. In each such 
            dictionary, every node in the graph G is a key. The value
            corresponding to each such key is a list. For each of the nrSamples
            created random networks, the value of the measure (specified by the
            outermost dictionary) at the node (specified by the next to outermost
            dictionary) is stored in this list. The length of the list of nrSamples.
    """
    #create the structure of embedded lists and dictionaries to store the data in:
    null_data = initialize_measures(G,measures) 
    #transform G into an unweighted multidigraph:
    G_multi = weightedDigraph_to_unweightedMultidigraph(G)
    #create the data:
    for i in range(nrSamples):
        print(i)
        F_multi=randomize(G_multi, self_loops,labeled) #produce a random sample
        F_weighted=unweightedMultidigraph_to_weightedDigraph(F_multi) #transform sample to weighted digraph
        for measure in measures:
            if measure == 'betweenness_centrality':
                null_data = compute_betweenness_centrality(F_weighted,null_data) #apply measure and add to null_data structure
            elif measure == 'clustering':
                null_data = compute_clustering(F_weighted,null_data) #apply measure and add to null_data structure
    return null_data

def initialize_measures(G,measures,data={}):
    """
    Function to create the structure of embedded lists and dictionaries to store the data in.
    
    input:
        1) weighted digraph G
        2) list of measures to use for the null model (list of strings)
        
    returns:
        dictionary. Every measure in the list 'measures' is a key. The value
            corresponding to each such key is a dictionary. In each such 
            dictionary, every node in the graph G is a key. The value
            corresponding to each such key is an empty list.
    """
    
    def recursive(measures,data):
        """ 
        recursively moves through the 'measures' list to create the desired structure.
        """ 
        if measures == []: #no measure to be added to the structure
            return data
        else:
            measure = measures[0] #pick a measure to add to the structure
            if measure == 'betweenness_centrality' or measure == 'clustering':
                data[measure] = {} #create key for outermost dictionary. Value is empty dictionary
                for node in G.nodes: #every node in G becomes a key to this new dictionary
                    data[measure][node]=[] #the values are empty lists.
            else:
                print('ERROR unknown measure: ', measures[0]) #measure not implemented
            measures.pop(0)
            return recursive(measures,data)
        
    tmp_measures = copy.deepcopy(measures) #create copy to not change original list
    return recursive(tmp_measures,data)
            

def compute_betweenness_centrality(G, data):
    """
    function that computes the betweenness centrality of each node in
        a multidigraph G and adds it to the correct list embedded in 'data'. 
        
    input: 
        1) unweighted multidigraph G
        2) data structure: dictionary with values that are dictionaries with values
                            that are lists. (see function initionalize_measures)
    returns: 
        data structure, with the betweenness centrality value for every node
            of G added to the correct lists embedded in 'data'.
    """
    values = betweenness_centrality(G) #compute betweenness centrality for every node of G
    for node in G.nodes:
        data['betweenness_centrality'][node].append(values[node]) #add values to 'data'
    return data

def compute_clustering(G, data):
    """
    function that computes the clustering coefficient of each node in
        a multidigraph G and adds it to the correct list embedded in 'data'. 
        
    input: 
        1) unweighted multidigraph G
        2) data structure: dictionary with values that are dictionaries with values
                            that are lists. (see function initionalize_measures)
    returns: 
        data structure, with the clustering coefficient value for every node
            of G added to the correct lists embedded in 'data'.
    """
    values = clustering_local(G) #compute clustering coefficient for every node of G
    for node in G.nodes:
        data['clustering'][node].append(values[node]) #add values to 'data'
    return data

       
def weightedDigraph_to_unweightedMultidigraph(G):
    """
    transforms a weighted digraph into an unweighted multidigraph. An edge of
        weight w is mapped to w copies of the edge in the multidigraph.
    input: weighted digraph G, with weights having key 'weight'
    returns: unweighted multidigraph

    """
    multiedges=[] #will be the edge list of the unweighted multidigraph
    for (tail,head,weight) in G.edges(data=True): #the edges of G
        for i in range(weight['weight']): 
            multiedges.append((tail,head)) #add the edge to the edge list of the output, multiplicity=weight
    H = nx.MultiDiGraph(multiedges) #construct multidigraph with new edge list
    return H

def unweightedMultidigraph_to_weightedDigraph(G):
    """
    transforms an unweighted multidigraph to a weighted digraph. w copies of an 
        edge in the multidigraph are mapped to an edge of weight w.
    input: unweighted multidigraph G
    returns: weighted digraph G, with weights having key 'weight'
    """
    H = nx.DiGraph() #will be the weighted digraph
    for (tail,head) in G.edges(data=False): #the edges of G
        if (tail,head) in H.edges: #edge has already been added to H
            H[tail][head]['weight']+=1 #add 1 to the weight of this edge
        else: #edge has not yet been added to H
            H.add_edge(tail,head,weight=1) #add edge with weight 1
    return H
    
